Unit Midi;

Interface

uses
 MidiFile,MidiFM,MidiSPK,Timer;

Type
 TDriver =(drNone,drSpeaker,drAdlib);
 TDrivers=set of TDriver;

Const
 drAll=[drSpeaker,drAdlib];

function Init(Drivers:TDrivers):boolean;
function Play(FileName:string):boolean;
function Playing:boolean;
procedure Stop;

Var
 Instrs:array[0..15] of byte;

{ midi driver procs }
var
 Driver   :TDriver;
 Reset    :procedure;
 Done     :procedure;
 NoteOn   :procedure(Channel,Note,Velocity:word);
 NoteOff  :procedure(Channel,Note,Velocity:word);
 SetVolume:procedure(Channel,Note,Velocity:word);

Implementation

var
 fPlaying:boolean;
 OldExit:pointer;

 T       :integer;
 Event   :byte;
 Channel :byte;
 Note    :byte;
 Velocity:byte;
 Pitch   :word;
 Meta    :byte;
 MetaSize:word;
 b       :byte;
 Time    :longint;

procedure Stop;
 begin
  if not fPlaying then exit;
  ExitProc:=OldExit;
  StopTimer;
  Reset;
  MidiFile.Close;
 end;

Function DelayCount:longint;
 Var
  Count:longint;
  t:integer;
 begin
  t:=1;
  Count:=0;
  while t<=TrackCount do begin
   Count:=Tracks^[t].Timer;
   if Count>=0 then break;
   inc(t);
  end;
  while t<=TrackCount do begin
   with Tracks^[T] do if (Timer>=0)and(Timer<Count) then Count:=Timer;
   inc(t);
  end;
  DelayCount:=Count;
 end;

Procedure TimerProc; far;
 Const
  Wait:longint=0;
  dly :longint=0;
 var
  Done:boolean;
  ms:integer;
 begin
  if Wait=0 then Wait:=DelayCount;
  if dly<=0 then Dly:=((Wait*Tempo) div Ticks);

  dec(dly,10); if dly>0 then exit;

  Done:=True;
  for T:=1 to MidiFile.TrackCount do

   With MidiFile.Tracks^[T] do begin

     if Timer<>-32000 then begin

      Done:=False;
      if Timer>0 then dec(Timer,Wait);

      if Timer=0 then begin

      Event:=Data^[Index];
      if Event<$80 then Event:=Last else begin Last:=Event; Inc(Index); end;

      Case Event shr 4 of

       $8: begin {Note Off }
            Channel :=Event and $F;
            Note    :=Data^[Index]; inc(index);
            Velocity:=Data^[Index]; inc(index);
            NoteOff(Channel,Note,Velocity);
           end;

       $9: begin {Note On }
            Channel :=Event and $F;
            Note    :=Data^[Index]; inc(index);
            Velocity:=Data^[Index]; inc(index);
            if Velocity=0 then
             NoteOff(Channel,Note,Velocity)
            else
             NoteOn(Channel,Note,Velocity);
           end;

       $A: begin { Key Aftertouch - only supported for GM device }
            Channel :=Event and $F;
            Note    :=Data^[Index]; inc(index);
            Velocity:=Data^[Index]; inc(index);
            SetVolume(Channel,Note,Velocity);
           end;

       $B: begin { Control change - only supported for GM device }
            Inc(Index,2);
           end;

       $C: begin { Patch change - this changes the instrument on a channel }
            Channel:=Event and $F;
            Instrs[Channel]:=Data^[Index]; inc(Index);
           end;

       $D: begin { Channel aftertouch - only supported on GM device }
            Inc(Index);
           end;

       $E: begin { Pitch wheel change - only supported on GM device }
            Inc(Index,2);
           end;

       $F: begin
            Meta:=Data^[Index]; inc(Index);

            MetaSize:=0;
            Repeat
             b:=Data^[Index]; inc(index);
             MetaSize:=MetaSize shl 7 + (b and $7f);
            Until b<128;

            Case Meta of
             $2F: Index:=Size;
             $51: Tempo:=(longint(Data^[Index+0]) shl 16
                         +word(Data^[Index+1]) shl 8
                         +Data^[Index+2]
                         ) div 1000;
            end;
           Inc(Index,MetaSize);
          end;

      end;

      if Index>=Size then
       Timer:=-32000
      else begin
       Done:=False;
       Time:=0;
       Repeat
        b:=Data^[Index]; Inc(Index);
        Time:=Time shl 7 + (b and $7F);
       Until b<128;
       Timer:=Time{Timer+(Time*Tempo) div Ticks};
      end;

     end; { Timer>0 }

    end; { Timer<>-1000 }

   end; { With Track^[T] }

   Wait:=DelayCount;

   if Done then Stop;
 end;


function Play(FileName:string):boolean;
 begin
  Play:=False;
  if Driver=drNone then exit;
  Stop;
  if not MidiFile.Open(FileName) then exit;
  fPlaying:=True;
  OldExit:=ExitProc;
  ExitProc:=@Stop;
  OnTimer:=TimerProc;
  StartTimer;
  Play:=True;
 end;

function Playing:boolean;
 begin
  Playing:=fPlaying;
 end;

var
 OldExit1:pointer;

procedure NewExit; far;
 begin
  ExitProc:=OldExit1;
  Done;
 end;

Function Init(Drivers:TDrivers):boolean;
 begin
 fPlaying:=False;
 if (drAdlib in Drivers) and MidiFM.Init then begin
  Driver   :=drAdlib;
  Reset    :=MidiFM.Reset;
  Done     :=MidiFM.Done;
  NoteOn   :=MidiFM.NoteOn;
  NoteOff  :=MidiFM.NoteOff;
  SetVolume:=MidiFM.SetVolume;
 end else
 if (drSpeaker in Drivers) then begin
  Driver   :=drSpeaker;
  Reset    :=MidiSPK.Reset;
  Done     :=MidiSPK.Done;
  NoteOn   :=MidiSPK.NoteOn;
  NoteOff  :=MidiSPK.NoteOff;
  SetVolume:=MidiSPK.SetVolume;
 end else begin
  Driver:=drNone;
  Init:=False;
  exit;
 end;
 Init:=True;
 OldExit1:=ExitProc;
 ExitProc:=@NewExit;
 Reset;
end;

end.