Unit MidiFile;

Interface

Function Open(FileName:string):boolean;
Procedure Close;

Procedure TrackInfo;

Type
 PByte=^TByte;
 TByte=array[0..65532] of byte;

 TChunk=record
  Sign:Array[1..4] of char;
  Size:record Hi:Word; Lo:Word end;
 end;

 THeader=record
  SMF   :word;
  Tracks:word;
  Times :word;
 end;

 TTrack=record
  Size  :word;
  Data  :PByte;
  Index :word;
  Last  :byte;
  Timer :longint;
  tTicks:word;
 {
  Voice :byte;
  Note  :byte;
  Octave:byte;
  Volume:byte;
  Instr :byte;
 }
 end;

 TTracks=array[1..100] of TTrack;

Var
 Ticks:word;
 Tempo:longint;

 TrackCount:word;
 Tracks:^TTracks;

Implementation

var
 MID:file;

 MIDIError:byte;
 { 0=No Error       }
 { 1=File not found }
 { 2=Invalid File   }
 { 3=Invalid Chunk  }

 T:Word;

 Chunk:TChunk;
 ChunkSize:word;
 ChunkData:PByte;
 ChunkIndex:word;


Function ReadChunk(Sign:string):boolean;
 begin
  BlockRead(MID,Chunk,SizeOf(Chunk));
  ReadChunk:=(Chunk.Sign=Sign)and(Chunk.Size.Hi=0);
  ChunkSize:=Swap(Chunk.Size.Lo);
 end;

Procedure GetChunk;
 begin
  GetMem(ChunkData,ChunkSize);
  BlockRead(MID,ChunkData^,ChunkSize);
  ChunkIndex:=0;
 end;

Procedure FreeChunk;
 begin
  FreeMem(ChunkData,ChunkSize);
 end;

Function ChunkTimer:longint;
 var
  Timer:longint;
  b:byte;
 begin
  Timer:=0;
  Repeat
   b:=ChunkData^[ChunkIndex]; inc(ChunkIndex);
   Timer:=Timer shl 7 + (b and $7F);
  Until b<128;
  ChunkTimer:=Timer;
 end;

Function ReadMidi:boolean;
 begin
  ReadMidi:=False;
  TrackCount:=0;

 { MIDI Header }
  MIDIError:=2; { invalid file }

  if not ReadChunk('MThd') then exit;
  GetChunk;
  With THeader(pointer(ChunkData)^) do begin
  { if Swap(SMF)=2 then exit;  { unsupported SMF=2 }
   Ticks:=Swap(Times);
   Tempo:=600;
   TrackCount:=Swap(Tracks);
  end;
  FreeChunk;

 { Tracks }
  MIDIError:=3; { Invalide Chunk }

  GetMem(Tracks,TrackCount*SizeOf(TTrack));
  FillChar(Tracks^,TrackCount*SizeOf(TTrack),0);

  For T:=1 to TrackCount do begin
    if not ReadChunk('MTrk') then Exit;
    With Tracks^[T] do begin
     GetChunk;
     Size  :=ChunkSize;
     Data  :=ChunkData;
     Timer :=ChunkTimer;
     Index :=ChunkIndex; { after Timer }
     Last  :=$FF;
    {
     Note  :=0;
     Octave:=0;
     Instr :=0;
     Volume:=0;
    }
    end;
  end;

  ReadMidi:=True;
 end;

Function Open(FileName:string):boolean;
 begin
  Open:=False;
  MIDIError:=1; { file not found }
  assign(MID,FileName);
  reset(MID,1);
  if IoResult<>0 then exit;
  Open:=ReadMidi;
  System.Close(MID);
 end;

Procedure Close;
 begin
  if TrackCount>0 then begin
   For T:=1 to TrackCount do
    with Tracks^[T] do if Size>0 then FreeMem(Data,Size);
   FreeMem(Tracks,TrackCount*SizeOf(TTrack));
   TrackCount:=0;
  end;
 end;

Procedure TrackInfo;
 begin
  For T:=1 to TrackCount do
   with Tracks^[T] do begin
    WriteLn('Track #',T:3,', ',Size:5,' bytes ',Timer,'s');
   end;
 end;

begin
 TrackCount:=0;
end.