Unit MidiFm;

interface

function  Init:boolean;
Procedure Reset;
procedure Done;

Procedure Delay(ms:word); { from the SWAG }

procedure NoteOff(Channel,Note,Velocity:word);
procedure NoteOn (Channel,Note,Velocity:word);

procedure SetVolume(Channel,Note,Velocity:word);

Var
 FMBase:word;     { $220, $240 or $388 }

implementation

uses Midi,AdlibGM;

Const
 AdlibFM=$388;

 NoSound=63;

 Op:array[0..8] of word = ( { i mod 3 + (i div 3) shl 3 }
  $000,$001,$002,$008,$009,$00A,$010,$011,$012
 );

 Notes:array[0..11] of word = (
  $157,$16B,$181,$198,$1B0,$1CA,$1E5,$202,$220,$241,$263,$287{,$2AE}
 );

 DrumChannels=[9]; { 10 ? 9..10 ? ????? }

 PercussiveVoice:array[0..4] of byte=(6,7,8,8,7);
 PercussiveFlags:array[6..10] of byte=($10,$08,$04,$02,$01);

type
 Instrument=record
  Sound   :word;
  iLevel  :word;
  ATDC    :word;
  STRL    :word;
  Wave    :word;
  Feedback:Byte;
 end;

var
{ NoteOn info }
 Channels:array[0..15] of record
           Instr:byte;
           Voice:array[0..127] of byte;
          end;

{ SBVoice Info }
 Voices  :array[0..8] of record
           Channel:byte;
           Note   :byte;
           Time   :longint;
          end;

 Percussive:byte;

 Time:longint;

Procedure delay(ms : word); assembler; { SWAG }
{ms is the number of milliseconds to delay. 1000ms = 1second}
 asm
  mov ax,1000
  mul ms
  jz  @done
  mov cx,dx
  mov dx,ax
  mov ah,$86
  int $15
  @done:
 end;

Procedure WriteFM(reg,val:word); assembler;
 asm
  mov dx,AdlibFM; mov ax,reg; out dx,al
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  inc dx; mov ax,Val; out dx,al
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
  in al,dx; in al,dx; in al,dx; in al,dx; in al,dx; in al,dx
 end;

Function ReadFM:byte; assembler;
 asm
  mov dx,AdlibFM
  in al,dx
 end;

Function Init:boolean;
 var
  s1,s2:byte;
 begin
  WriteFM($04,$60);   (* reset both timers        *)
  WriteFM($04,$80);   (* enable timer interrupts  *)
  s1:=ReadFM;         (* read status register     *)
  WriteFM($02,$FF);
  WriteFM($04,$21);   (* start timer 1            *)
  Delay(80);          (* could do something useful*)
  s2:=ReadFM;         (* read status register     *)
  WriteFM($04,$60);   (* reset both timers        *)
  WriteFM($04,$80);   (* enable timer interrupts  *)

  Init:=(((s1 and $E0)=0)and((s2 and $E0)=$C0));

  fillchar(Channels,SizeOf(Channels),255);
  fillchar(Voices,SizeOf(Voices),255);
 end;

Procedure Reset;
 var
  r:byte;
 begin
  (* zero all registers *)
  for r:=$01 to $F5 do WriteFM(r,0);
  (* allow FM chips to control the waveform of each operator *)
  WriteFM($01,$20);

{  WriteFM($08,(1 shl 6)); }
  (* set rhythm enabled (6 melodic voices, 5 percussive) *)
  Percussive:=(0
             {  (1 shl 7) { Amp }
             { +(1 shl 6) { Vibrato }
             { +(1 shl 5)  { rhythm }
             { +(1 shl 4) { bass }
             { +(1 shl 3) { snare }
             { +(1 shl 2) { tom-tom }
             { +(1 shl 1) { cymbal }
             { +(1)       { hithat }
              );
{  WriteFM($BD,Percussive); }
 end;

procedure SetInstrument(Drum:boolean;Voice,IBKIndex,Velocity:word);
 var
  Operator:word;
  Level:word;
  Instr:^Instrument;
 begin
  Operator:=Op[Voice];

  if Drum then begin
   if IBKIndex<5 then IBKIndex:=0 else begin
    dec(IBKIndex,5);
    if IBKIndex>46 then IBKIndex:=46;
   end;
   Instr:=@Drums[IBKIndex]
  end else
   Instr:=@Melodic[IBKIndex];

  With Instr^ do begin

   WriteFM(Operator+$20,Lo(Sound));
   WriteFM(Operator+$60,Lo(ATDC));
   WriteFM(Operator+$80,Lo(STRL));

   if not Drum then begin
    WriteFM(Operator+$23,Hi(Sound));
    WriteFM(Operator+$63,Hi(ATDC));
    WriteFM(Operator+$83,Hi(STRL));
   end;

  {
   WriteFM(Operator+$E0,Lo(Wave));
   WriteFM(Operator+$E3,Hi(Wave));
  }

   WriteFM(Voice+$C0,FeedBack);

   Level:=
    (Hi(iLevel) and $C0 + 63-(((63-Hi(iLevel) and $3F)*Velocity) div 127)) shl 8
   +(Lo(iLevel) and $C0 + 63-(((63-Lo(iLevel) and $3F)*Velocity) div 127));

   if Drum then begin
    if Voice=6 then begin
     WriteFM(Operator+$43,Hi(Level))
    end else begin
     WriteFM(Operator+$40,Lo(Level));
    end;
   end else begin
    if Boolean(FeedBack and 1) then
     WriteFM(Operator+$43,Hi(Level))
    else
     WriteFM(Operator+$43,Hi(iLevel));
    WriteFM(Operator+$40,Lo(Level));
   end;

  end;
 end;

procedure NoteOff(Channel,Note,Velocity:word);
 Var
  Voice:byte;
  N,O:word;
 begin
  Voice:=Channels[Channel].Voice[Note];
  if Voice=255 then exit;

  Channels[Channel].Voice[Note]:=255;
  Voices[Voice].Channel:=255;

  if Channel in DrumChannels then begin
   Percussive:=Percussive and NOT (PercussiveFlags[Instrs[Channel]]);
   WriteFM($BD,Percussive);
  end else begin

   N:=Notes[Note mod 12];
   O:=Note div 12;

   WriteFM(Voice+$A0,Lo(N));
   WriteFM(Voice+$B0,hi(N) or (O shl 2));
  end;
 end;

procedure NoteOn(Channel,Note,Velocity:word);
 Var
  Drum:boolean;
  Voice:byte;
  Oldest:longint;
  N,O:word;
 begin
  Drum:=(Channel in DrumChannels) and (Instrs[Channel]<5);

  If Drum then begin

   Voice:=PercussiveVoice[Instrs[Channel]];

  end else begin

   Voice:=0;
   Oldest:=Voices[Voice].Time;
   For N:=0 to 5 do begin
    if Voices[N].Channel=255 then begin
     Voice:=N;
     break;
    end;
    if Voices[N].Time<Oldest then begin
     Voice:=N;
     Oldest:=Voices[N].Time;
    end;
   end;

  end;

  With Voices[Voice] do if Channel<255 then NoteOff(Channel,Note,0);

  Channels[Channel].Voice[Note]:=Voice;

  Voices[Voice].Channel:=Channel;
  Voices[Voice].Note:=Note;
  inc(Time);
  Voices[Voice].Time:=Time;

  SetInstrument(Drum,Voice,Instrs[Channel],Velocity);

  N:=Notes[Note mod 12];
  O:=Note div 12;

  WriteFM(Voice+$A0,lo(N));
  if Drum then begin
   Percussive:=Percussive or PercussiveFlags[Instrs[Channel]];
   WriteFM($BD,Percussive);
   WriteFM(Voice+$B0,hi(N) or (O shl 2));
  end else
   WriteFM(Voice+$B0,hi(N) or (O shl 2) or $20);
 end;

Procedure SetVolume(Channel,Note,Velocity:word);
 Var
  Voice:byte;
  Drum:boolean;
 begin
  Voice:=Channels[Channel].Voice[Note];
  if Voice=255 then exit;
  Drum:=(Channel in DrumChannels) and (Instrs[Channel]<5);
  SetInstrument(Drum,Voice,Instrs[Channel],Velocity);
 end;

Procedure Done;
 var
  i:integer;
 begin
  Reset;
  for i:=0 to 15 do WriteFM($40+i,63);
  for i:=0 to 15 do WriteFM($B0+i,0);
 end;

end.
